/*
  isp.c - part of USBasp

  Autor..........: Thomas Fischl <tfischl@gmx.de> 
  Description....: Provides functions for communication/programming
                   over ISP interface
  Licence........: Free under certain conditions. See Documentation.
  Creation Date..: 2005-02-23
  Last change....: 2005-04-20
*/
/*
 * 2006-05-06 Target ATMega48/ATMega88/ATMega168 modfied by t.kuroki
 * 2006-05-08 append ispSetSCKOptionEx() function
 * 2006-05-13 uchar ispTransmit_sw(uchar send_byte) modified.
 * 2006-05-18 hand optimized
 * 2006-05-20 support for mega128x/mega256x
 */

#include <avr/io.h>
#include "isp.h"
#include "clock.h"

#define spiHWdisable() SPCR = 0

static uchar spcr_value;		//@@+ by t.k
static uchar spsr_value;		//@@+ by t.k
static uchar isp_delay_value;	//@@+ by t.k

//

void spiHWenable(void)
{
  /* enable SPI, master, 375kHz SCK */
//  SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR1);
//  SPSR = (1 << SPI2X);
	SPCR = spcr_value;	//@@+ by t.k
	SPSR = spsr_value;	//@@+ by t.k
}

/* by t.kuroki
 *   fck  < 12MHz : Low period>2CPU clock cycle   High period>2CPU clock cycle
 *   fck >= 12MHz : Low period>3CPU clock cycle   High period>3CPU clock cycle
 * 
 *       SCK clock           cycle      TARGET
 *   0 : 3MHz      fosc/4    0.33us    > 12MHz
 *   1 : 1.5MHz    fosc/8    0.67us    > 6MHz
 *   2 : 750KHz    fosc/16   1.33us    > 3MHz
 * @ 3 : 375KHz    fosc/32   2.67us    > 1.5MHz
 *   4 : 187.5KHz  fosc/64   5.33us    > 750KHz
 *   5 : 93.75KHz  fosc/128 10.67us    > 375KHz
 * 
 *   6 :        (6-5)*10*2     20us    > 200KHz
 *   7 :        (7-5)*10*2     40us    > 100KHz
 *   8 :        (8-5)*10*2     60us    > 66.7KHz
 *   9 :        (9-5)*10*2     80us    > 50KHz
 *  10 :       (10-5)*10*2    100us    > 40KHz
 *  11 :       (11-5)*10*2    120us    > 33.3KHz
 * 
 * 255 :      (255-5)*10*2      5ms    > 800Hz
 */
void ispSetSCKOption(uchar option)		//@@x
{
	isp_delay_value = 6;				//@@ 60us by t.k
	if (option < 6)
	{
		/* use hardware spi */
		set_hw_spi();		//@@x ispTransmit = ispTransmit_hw;
		if (option == 0)
		{	// fosc/4
			spcr_value = (1 << SPE) | (1 << MSTR);
			spsr_value = (0 << SPI2X);
		}
		else if (option == 1)
		{	// fosc/8
			spcr_value = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			spsr_value = (1 << SPI2X);
		}
		else if (option == 2)
		{	// fosc/16
			spcr_value = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			spsr_value = (0 << SPI2X);
		}
		else if (option == 3)
		{	// fosc/32 == original High-speed
			spcr_value = (1 << SPE) | (1 << MSTR) | (1 << SPR1);
			spsr_value = (1 << SPI2X);
		}
		else if (option == 4)
		{	// fosc/64
			spcr_value = (1 << SPE) | (1 << MSTR) | (1 << SPR1);
			spsr_value = (0 << SPI2X);
		}
		else
		{	// fosc/128
			spcr_value = (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0);
			spsr_value = (0 << SPI2X);
		}
	}
	else
	{
		/* use software spi */
		set_sw_spi();	//@@x ispTransmit = ispTransmit_sw;
		isp_delay_value = option-5;			//@@ (option-5)*10us by t.k
	}
}

// original ispDelay() was 64us wait
#define ispDelay()	delay_10us(isp_delay_value)	//@@+ by t.k

void ispConnect(void)
{
	/* all ISP pins are inputs before */
	/* now set output pins */
	ISP_DDR |= (1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI);

	/* reset device */
	ISP_OUT &= ~(1 << ISP_RST);		/* RST low */
	ISP_OUT &= ~(1 << ISP_SCK);		/* SCK low */

	/* positive reset pulse > 2 SCK (target) */
	delay_10us(6);					//@@x  ispDelay();
	ISP_OUT |= (1 << ISP_RST);		/* RST high */
	delay_10us(100);				//@@x  ispDelay(); -> 1ms
	ISP_OUT &= ~(1 << ISP_RST);		/* RST low */

	delay_100us(250);				//@@+ 25ms >= 20ms by t.k

	if (is_hw_spi())
	{
		spiHWenable();
	}
}

void ispDisconnect(void)
{
	/* set all ISP pins inputs */
	ISP_DDR &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI));
	/* switch pullups off */
	ISP_OUT &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI));

	/* disable hardware SPI */
	spiHWdisable();
}

uchar ispTransmit(uchar send_byte)				//@@@ by t.k
{
	if (is_hw_spi())
	{
		SPDR = send_byte;
		while (!(SPSR & (1 << SPIF)))
			;
		return SPDR;
	}
	else
	{
		uchar rec_byte = 0;
		uchar n = 8;
		do
		{
			/* set MSB to MOSI-pin */
			if ((send_byte & 0x80) != 0)
			{
				ISP_OUT |= (1 << ISP_MOSI);  /* MOSI high */
			}
			else
			{
				ISP_OUT &= ~(1 << ISP_MOSI); /* MOSI low */
			}
			/* shift to next bit */
			send_byte  = send_byte << 1;

			ispDelay();			//@@+ MOSI setup time 2006-05-13 added by t.k

			/* receive data */
			rec_byte = rec_byte << 1;
			if ((ISP_IN & (1 << ISP_MISO)) != 0)
			{
				rec_byte++;
			}

			/* pulse SCK */
			ISP_OUT |= (1 << ISP_SCK);     /* SCK high */
			ispDelay();
			ISP_OUT &= ~(1 << ISP_SCK);    /* SCK low */
	//@@-    ispDelay();
		} while (--n);

		return rec_byte;
	}
}

static void __inline__ ispTransmitWord(uint16_t w)	//@@+
{
	union {
		uint8_t b[2];
		uint16_t w;
	} u;
	u.w = (w);
	ispTransmit(u.b[1]);
	ispTransmit(u.b[0]);
}

//@@+ for mega128x/mega256x by t.k
static void __inline__ ispTransmitFlashAddr(uint16_t addr)
{
	union {
		uint8_t b[2];
		uint16_t w;
	} u;
	u.w = addr >> 1;
	if (tst_addr16())
		u.b[1] |= 0x80;
	ispTransmit(u.b[1]);
	ispTransmit(u.b[0]);
}

//@@+ for mega256x by t.k
void ispLoadExtendedAddress(uchar eaddr)
{
	ispTransmit(0x4D);			/* extended address command */
	ispTransmit(0x00);
	ispTransmit(eaddr);			/* address extended */
	ispTransmit(0x00);
}

uchar ispEnterProgrammingMode(void)
{
	uchar check;
	uchar count = 32;

	do
	{
		ispTransmit(0xAC);
		ispTransmit(0x53);
		check = ispTransmit(0);
		ispTransmit(0);

		if (check == 0x53)
		{
			return 0;
		}

		spiHWdisable();

		delay_10us(6);					//@@+ by t.k
		/* pulse SCK */
		ISP_OUT |= (1 << ISP_SCK);		/* SCK high */
		delay_10us(6);					//@@x    ispDelay();
		ISP_OUT &= ~(1 << ISP_SCK);		/* SCK low */
		delay_10us(6);					//@@x    ispDelay();

		if (is_hw_spi())
		{
			spiHWenable();
		}

	} while (--count);

	return 1;  /* error: device dosn't answer */
}

uchar ispReadFlash(unsigned int address)
{
	uchar cmd = 0x20;
	if ((uchar)(address & 1))
		cmd = 0x28;
	ispTransmit(cmd);
	ispTransmitFlashAddr(address);
	return ispTransmit(0);
}


uchar ispWriteFlash(unsigned int address, uchar data, uchar pollmode)
{
	/* 0xFF is value after chip erase, so skip programming 
	if (data == 0xFF) {
		return 0;
	}
	*/

	uchar cmd = 0x40;
	if ((uchar)(address & 1))
		cmd = 0x48;
	ispTransmit(cmd);
	ispTransmitFlashAddr(address);
	ispTransmit(data);

	if (pollmode == 0)
		return 0;

	if (data == 0x7F)
	{
		delay_100us(48);	//@@x clockWait(15); /* wait 4,8 ms */
		return 0;
	}
	else
	{
		/* polling flash */
		uchar retries = 30;		// 30*340us = 9.6ms
		uint8_t starttime = TIMERVALUE;
		do
		{
			if (ispReadFlash(address) != 0x7F)
			{
				return 0;
			}

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us)
			{
				starttime += CLOCK_T_320us;		//@@ by t.kuroki
				retries --;
			}

		} while (retries != 0);
		return 1; /* error */
	}
}


uchar ispFlushPage(unsigned int address, uchar pollvalue)
{
	ispTransmit(0x4C);
	ispTransmitFlashAddr(address);
	ispTransmit(0);


	if (pollvalue == 0xFF)
	{
		delay_100us(48);	//@@x clockWait(15); 4.8ms
		return 0;
	}
	else
	{
		/* polling flash */
		uchar retries = 30;		// 30*340us = 9.6ms
		uint8_t starttime = TIMERVALUE;
		do
		{
			if (ispReadFlash(address) != 0xFF)
			{
				return 0;
			}

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us)
			{
				starttime += CLOCK_T_320us;		//@ by t.kuroki
				retries --;
			}

		} while (retries != 0);
		return 1; /* error */
	}
 }


uchar ispReadEEPROM(unsigned int address)
{
	ispTransmit(0xA0);
	ispTransmitWord(address);
	return ispTransmit(0);
}


uchar ispWriteEEPROM(unsigned int address, uchar data)
{
	ispTransmit(0xC0);
	ispTransmitWord(address);
	ispTransmit(data);

	delay_100us(96);	//@@x  clockWait(30); // wait 9,6 ms 

	return 0;
	/* 
	if (data == 0xFF)
	{
		delay_100us(96);	//@@x  clockWait(30); // wait 9,6 ms 
		return 0;
	}
	else
	{
		// polling eeprom 
		uchar retries = 30; // about 9,6 ms 
		uint8_t starttime = TIMERVALUE;
		do
		{
			if (ispReadEEPROM(address) != 0xFF) {
				return 0;
			}

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us)
			{
				starttime = TIMERVALUE;
				retries --;
			}

		} while (retries != 0);
		return 1; // error 
	}
	*/

}
