/*
  USBasp - USB in-circuit programmer for Atmel AVR controllers

  Thomas Fischl <tfischl@gmx.de>

  License:
  The project is built with AVR USB driver by Objective Development, which is
  published under an own licence based on the GNU General Public License (GPL).
  USBasp is also distributed under this enhanced licence. See Documentation.

  Target.........: ATMega8 at 12 MHz
  Creation Date..: 2005-02-20
  Last change....: 2005-10-08

  PC2 SCK speed option. GND  -> slow (8khz SCK),
                        open -> fast (375kHz SCK)
*/
/*
 * 2006-05-06 Target ATMega48/ATMega88/ATMega168 modfied by t.kuroki
 * 2006-05-08 append ispSetSCKOptionEx() function
 * 2006-05-18 Rev 1.11 hand optimized
 * 2006-05-18 Rev 1.12 USBASP_FUNC_CONNECTT return-value
 * 2006-05-20 support for mega128x/mega256x
 * 2006-05-21 re-connect after serial-number changed
 */

#include <stdint.h>			//@@+ by t.k
#include <stdbool.h>		//@@+ by t.k
#include <avr/io.h>
//#include <avr/signal.h>	//@@- obsolete
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>		//@@+ by t.k

#include "usbdrv.h"
#include "isp.h"
#include "clock.h"

#define USBASP_FUNC_CONNECT		1
#define USBASP_FUNC_DISCONNECT	2
#define USBASP_FUNC_TRANSMIT	3
#define USBASP_FUNC_READFLASH	4
#define USBASP_FUNC_ENABLEPROG	5
#define USBASP_FUNC_WRITEFLASH	6
#define USBASP_FUNC_READEEPROM	7
#define USBASP_FUNC_WRITEEEPROM	8

#define USBASPX_FUNC_GETINFO	90
#define USBASPX_FUNC_SETSERIAL	91

#define PROG_STATE_IDLE			0
#define PROG_STATE_WRITEFLASH	1
#define PROG_STATE_READFLASH	2
#define PROG_STATE_READEEPROM	3
#define PROG_STATE_WRITEEEPROM	4

#define PROG_BLOCKFLAG_FIRST	0x01
#define PROG_BLOCKFLAG_LAST		0x02

#define PROG_FLAG_ADDR_16		0x04	//@@+ for mega128x/mega256x
#define PROG_FLAG_ADDR_17		0x08	//@@+ for mega256x
#define PROG_FLAG_ADDR_EX		0x80	//@@+ for mega256x

#define PROG_FLAGS_MASK			0xCF	//@@+
#define PROG_PAGESIZE_MASK		0x30	//@@+

#ifndef USBASPX_VERSION
//#define USBASPX_VERSION			0x100
#error version error
#endif
//----------------------------------------------------------------------------
#define INI_PORTB		0b00000000		// no pullups on USB and ISP pins
#define INI_DDRB		0b00000000		// all USB and ISP pins inputs
#define INI_PORTC		0b11111110
#define INI_DDRC		0b00000011		// all inputs except PC0, PC1
#define INI_PORTD		0b00000000
#define INI_DDRD		0b11111011		// all outputs except PD2 = INT0

#define LED_PORT		PORTC
#define LED_DDR			DDRC
#define LED_RED			PC1
#define LED_GREEN		PC0

#define ledRedOn()		LED_PORT &= ~(1 << LED_RED)
#define ledRedOff()		LED_PORT |= (1 << LED_RED)
#define ledGreenOn()	LED_PORT &= ~(1 << LED_GREEN)
#define ledGreenOff()	LED_PORT |= (1 << LED_GREEN)

//----------------------------------------------------------------------------
/*
 * #define USB_CFG_SERIAL_NUMBER_LENGTH  5
 * int usbCfgSerialNumberStringDescriptor[] PROGMEM = {
 *     USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH),
 *     '1', '2', '3', '4', '5'
 * };
 */
//@@+ ROM -> RAM by t.k
int usbCfgSerialNumberStringDescriptor[USB_CFG_SERIAL_NUMBER_LENGTH+1];

struct st_eeprom {
	uint8_t eeprom_reserved[16];
	int SerialNumber[USB_CFG_SERIAL_NUMBER_LENGTH+1];
};

struct st_eeprom eeprom EEMEM = {
	{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 },	// reserved.
	{
		USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH),
		'0', '0', '0', '0'
	}
};

void loadSerialNumber(void)
{
	EECR = 0;
	eeprom_read_block(usbCfgSerialNumberStringDescriptor,
		eeprom.SerialNumber, sizeof(usbCfgSerialNumberStringDescriptor));
	if (usbCfgSerialNumberStringDescriptor[0] !=
	   USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH))
	{
		usbCfgSerialNumberStringDescriptor[0] = USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH);
		usbCfgSerialNumberStringDescriptor[1] = '0';
		usbCfgSerialNumberStringDescriptor[2] = '0';
		usbCfgSerialNumberStringDescriptor[3] = '0';
		usbCfgSerialNumberStringDescriptor[4] = '0';
	}
}

void saveSerialNumber(void)
{
	eeprom_write_block(usbCfgSerialNumberStringDescriptor,
		eeprom.SerialNumber, sizeof(usbCfgSerialNumberStringDescriptor));
}
//----------------------------------------------------------------------------

static uchar replyBuffer[8];

static uchar prog_state = PROG_STATE_IDLE;

static unsigned int prog_address;
static unsigned int prog_nbytes = 0;
//static unsigned int prog_pagesize;	 //TP: Mega128 fix
static uchar prog_pagesize_lo;
static uchar prog_pagesize_hi;	 	//TP: Mega128 fix
//static uchar prog_blockflags;		//@@-
static uchar prog_flags;			//@@+
static uchar prog_pagecounter;

static bool f_reconnect;			//@@+ by t.k

#define GET_WORD(ptr)	(*(unsigned int *)(ptr))

//----------------------------------------------------------------------------

uchar usbFunctionSetup(uchar data[8])
{

	uchar len = 0;

	if (data[1] == USBASP_FUNC_READFLASH)
	{
		prog_address = GET_WORD(&data[2]);
		if (data[4] == 0x5A)		// magic no
			prog_flags = data[5] & PROG_FLAGS_MASK;	//@@+
		else
			prog_flags = 0;
		prog_nbytes  = GET_WORD(&data[6]);
	    prog_state   = PROG_STATE_READFLASH;
	    len          = 0xff; /* multiple in */
	}
	else if (data[1] == USBASP_FUNC_WRITEFLASH)
	{
		prog_address     = GET_WORD(&data[2]);
		prog_pagesize_lo = data[4];
	    prog_flags       = data[5] & PROG_FLAGS_MASK;	//@@x
		prog_pagesize_hi = (data[5] & PROG_PAGESIZE_MASK) >> 4;	//TP: Mega128 fix
		if (prog_flags & PROG_BLOCKFLAG_FIRST)
		{
			prog_pagecounter = prog_pagesize_lo;	// <- 0 = 256 bytes
	    }
		prog_nbytes = GET_WORD(&data[6]);
	    prog_state  = PROG_STATE_WRITEFLASH;
	    len         = 0xff; /* multiple out */
	}
	else if (data[1] == USBASP_FUNC_TRANSMIT)
	{
	    replyBuffer[0] = ispTransmit(data[2]);
	    replyBuffer[1] = ispTransmit(data[3]);
	    replyBuffer[2] = ispTransmit(data[4]);
	    replyBuffer[3] = ispTransmit(data[5]);
	    len            = 4;
	}
	else if (data[1] == USBASP_FUNC_READEEPROM)
	{
		prog_address = GET_WORD(&data[2]);
		prog_nbytes  = GET_WORD(&data[6]);
	    prog_state   = PROG_STATE_READEEPROM;
	    len          = 0xff; /* multiple in */
	}
	else if (data[1] == USBASP_FUNC_WRITEEEPROM)
	{
		prog_address     = GET_WORD(&data[2]);
		prog_pagesize_lo = prog_pagesize_hi = 0;
	    prog_flags       = 0;
		prog_nbytes      = GET_WORD(&data[6]);
	    prog_state       = PROG_STATE_WRITEEEPROM;
	    len              = 0xff; /* multiple out */
	}
	else if (data[1] == USBASP_FUNC_ENABLEPROG)
	{
	    replyBuffer[0] = ispEnterProgrammingMode();
	    len            = 1;
	}
	else if (data[1] == USBASP_FUNC_CONNECT)
	{
		if (data[3] == 0xaa && data[4] == 0x55)		//@@+ by t.k
		{
			/* set SCK speed */
			ispSetSCKOption(data[5]);
		}
		else
		{
			/* == original */
			/* set SCK speed */
			if ((PINC & (1 << PC2)) == 0) {
				ispSetSCKOption(ISP_SCK_SLOW);
			} else {
				ispSetSCKOption(ISP_SCK_FAST);
			}
		}
		ispConnect();
		ledRedOn();
		replyBuffer[0] = 0;		//@@+ by t.k
		len = 1;				//@@+ by t.k
	}
	else if (data[1] == USBASP_FUNC_DISCONNECT)
	{
    	ispDisconnect();
    	ledRedOff();
	}
	else if (data[1] == USBASPX_FUNC_GETINFO)		//@@+ by t.k
	{
	    replyBuffer[0] = (uint8_t)USBASPX_VERSION;
	    replyBuffer[1] = USBASPX_VERSION >> 8;
	    replyBuffer[2] = PINC;
	    replyBuffer[3] = 0;
	    len            = 4;
	}
	else if (data[1] == USBASPX_FUNC_SETSERIAL)		//@@+ by t.k
	{
		usbCfgSerialNumberStringDescriptor[0] = USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH);
		usbCfgSerialNumberStringDescriptor[1] = data[2];
		usbCfgSerialNumberStringDescriptor[2] = data[3];
		usbCfgSerialNumberStringDescriptor[3] = data[4];
		usbCfgSerialNumberStringDescriptor[4] = data[5];
		saveSerialNumber();
	    replyBuffer[0] = 0;
	    len = 1;
		f_reconnect = true;

	}

	usbMsgPtr = replyBuffer;

	return len;
}

//@@+ for mega128x/mega256x by t.k
void setupExtendedAddress(void)
{
	if (prog_flags & PROG_FLAG_ADDR_EX)		//@@+ for mega256x
	{
		if (prog_flags & PROG_FLAG_ADDR_17)
			ispLoadExtendedAddress(0x01);
		else
			ispLoadExtendedAddress(0x00);
	}
	if (prog_flags & PROG_FLAG_ADDR_16)		//@@+ for mega128x/mega256x
		set_addr16();
	else
		clr_addr16();
}

uchar usbFunctionRead(uchar *data, uchar len)
{
	uchar (*read_func)(unsigned int);

	/* check if programmer is in correct read state */
	if (prog_state == PROG_STATE_READFLASH)
	{
		read_func = ispReadFlash;
		setupExtendedAddress();		//@@+ for mega128x/mega256x
	}
	else if (prog_state == PROG_STATE_READEEPROM)
		read_func = ispReadEEPROM;
	else
		return 0xff;

	/* fill packet */
	uchar i = len;
	unsigned int addr = prog_address;
	do {
		*data++ = read_func(addr);
		addr++;
	} while (--i);
	prog_address = addr;

  /* last packet? */
	if (len < 8)
		prog_state = PROG_STATE_IDLE;

	return len;
}


uchar usbFunctionWrite(uchar *data, uchar len)
{
	unsigned int addr  = prog_address;
	unsigned int count = prog_nbytes;

	/* check if programmer is in correct write state */
	if (prog_state == PROG_STATE_WRITEEEPROM)
	{
		while (len)
		{
			/* EEPROM */
			ispWriteEEPROM(addr, *data);
			++data;

			if (!--count)
			{
				prog_nbytes = 0;
				prog_state  = PROG_STATE_IDLE;
				// final package (for new AVR-USB) 2006-05-07 by t.kuroki
				return 1;
			}

			++addr;
			--len;
		}
		prog_address = addr;
		prog_nbytes  = count;
		return 0;
	}
	else if (prog_state != PROG_STATE_WRITEFLASH)
		return 0xff;

	/* Flash */
	setupExtendedAddress();		//@@+ for mega128x/mega256x
	bool not_paged = (!prog_pagesize_lo && !prog_pagesize_hi);
	uchar page_counter = prog_pagecounter;
	while (len)
	{
		/* Flash */
		ispWriteFlash(addr, *data, not_paged);
		if (!not_paged)
		{
			/* paged */
			--page_counter;		/* 0 -> 256 */
			if (!page_counter)
			{
				ispFlushPage(addr, *data);
				page_counter = prog_pagesize_lo;
			}
		}

		if (!--count)
		{
			prog_nbytes      = 0;
			prog_pagecounter = page_counter;
			prog_state       = PROG_STATE_IDLE;
			if ((prog_flags & PROG_BLOCKFLAG_LAST) &&
				(page_counter != prog_pagesize_lo))
			{
				/* last block and page flush pending, so flush it now */
				ispFlushPage(addr, *data);
			}
			// final package (for new AVR-USB) 2006-05-07 by t.kuroki
			return 1;
		}

		++data;
		++addr;
		--len;
	}

	prog_pagecounter = page_counter;
	prog_address     = addr;
	prog_nbytes      = count;
	return 0;
}

#ifndef USB_CFG_PULLUP_IOPORT	//@@+ by t.k
#define usbDeviceConnect()		(USBDDR &= ~USBMASK)
#define usbDeviceDisconnect()	((USBOUT &= ~USBMASK),(USBDDR |= USBMASK))
#endif

void ReConnect(void)
{
	cli();
    USB_INTR_ENABLE &= ~(1 << USB_INTR_ENABLE_BIT);
	USB_INTR_PENDING = (1 << USB_INTR_PENDING_BIT);
	usbDeviceDisconnect();
	delay_100us(100);		// 10ms
	f_reconnect = false;
	usbDeviceConnect();
	usbInit();
	sei();
}

int main(void)
{
	PORTB = INI_PORTB;
	DDRB  = INI_DDRB;
	PORTC = INI_PORTC;
	DDRC  = INI_DDRC;
	PORTD = INI_PORTD;
	DDRD  = INI_DDRD;

	loadSerialNumber();

	clockInit();          /* init timer */

	ispSetSCKOption(ISP_SCK_FAST);
//@@-	usbInit();
//@@-	sei();
	ReConnect();
	do
	{	/* main event loop */
		usbPoll();

		if (f_reconnect)
		{
			uint8_t count = 156;	// 50ms
			uint8_t starttime = TIMERVALUE;
			do
			{
				usbPoll();
				if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us)
				{
					starttime += CLOCK_T_320us;
					--count;
				}

			} while (count != 0);
			ReConnect();
		}
	} while (true);

	return 0;
}

